import asyncio
import contextlib
import fcntl
import signal
import traceback

_RELAY_BUFFER_SIZE = 64 * 1024  # 64 KiB


def run(coro, debug=False):
    def signal_handler(task):
        task.cancel()

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        if debug is not None:
            loop.set_debug(debug)

        task = loop.create_task(coro)
        for signame in ["SIGHUP", "SIGINT", "SIGTERM"]:
            loop.add_signal_handler(getattr(signal, signame), signal_handler, task)

        loop.run_until_complete(task)
    except asyncio.CancelledError:
        pass
    finally:
        loop.run_until_complete(loop.shutdown_asyncgens())
        if hasattr(loop, "shutdown_default_executor"):
            loop.run_until_complete(loop.shutdown_default_executor())
        asyncio.set_event_loop(None)
        loop.close()


async def run_concurrently(*coros):
    """Run coros concurrently and cancel others on error.

    Like `asyncio.gather()`, but if one task raises an exception, all other
    tasks are cancelled.
    """

    tasks = [asyncio.create_task(coro) for coro in coros]
    try:
        return await asyncio.gather(*tasks)
    finally:
        await cancel_tasks(tasks)


@contextlib.asynccontextmanager
async def background_task(coro):
    """Runs the coro until leaving the context manager.

    The coro task is canceled when leaving the context."""
    task = asyncio.create_task(coro)
    try:
        yield task
    finally:
        await cancel_tasks([task])


async def cancel_tasks(tasks):
    """Cancel tasks and wait until all are cancelled"""

    canceled_tasks = []
    for task in tasks:
        if not task.done():
            task.cancel()
            canceled_tasks.append(task)

    for task in canceled_tasks:
        try:
            await task
        except asyncio.CancelledError:
            pass


@contextlib.asynccontextmanager
async def on_error(callback, /, *args, **kwargs):
    """Calls a cleanup callback, if an exception is raised within the
    context manager.
    """
    async with contextlib.AsyncExitStack() as stack:
        stack.push_async_callback(callback, *args, **kwargs)
        yield
        stack.pop_all()


@contextlib.asynccontextmanager
async def connect(*args, **kwargs):
    """Wraps `asyncio.open_connection()` in a context manager

    The connection is closed when leaving the context.
    """

    reader, writer = await asyncio.open_connection(*args, **kwargs)
    try:
        yield reader, writer
    finally:
        writer.close()
        await writer.wait_closed()


async def relay_streams(client_r, client_w, remote_r, remote_w):
    """Relay data between two streams"""

    async def relay(reader, writer):
        while True:
            data = await reader.read(_RELAY_BUFFER_SIZE)
            if not data:
                writer.write_eof()
                return

            writer.write(data)
            await writer.drain()

    await run_concurrently(relay(client_r, remote_w), relay(remote_r, client_w))


class Server:
    """Wraps `asyncio.start_server()` and cleans up open connections

    When leaving the context, not only are the listening sockets closed,
    but all currently running connection handler callbacks are also
    canceled.

    It also wraps the connection handler to catch and log every exception
    raised in the handler and also closes the connection, when the handler
    function returns.
    """

    def __init__(self, connection_handler, *args, **kwargs):
        self._connection_handler = connection_handler
        self._args = args
        self._kwargs = kwargs
        self._tasks = set()

    async def __aenter__(self):
        self._server = await self._start()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        self._server.close()
        await self._server.wait_closed()
        await cancel_tasks(self._tasks.copy())

    def _on_connect(self, reader, writer):
        task = asyncio.create_task(self._run_handler(reader, writer))
        self._tasks.add(task)
        task.add_done_callback(self._tasks.discard)

    async def _run_handler(self, reader, writer):
        try:
            await self._connection_handler(reader, writer)
        except Exception:
            traceback.print_exc()
        finally:
            try:
                writer.close()
                await writer.wait_closed()
            except Exception:
                pass

    async def _start(self):
        return await asyncio.start_server(self._on_connect, *self._args, **self._kwargs)

    async def serve_forever(self):
        await self._server.serve_forever()


class UnixServer(Server):
    """Same as `Server`, but for `asyncio.start_unix_server()`"""

    async def _start(self):
        return await asyncio.start_unix_server(
            self._on_connect, *self._args, **self._kwargs
        )


class ContextStack:
    """Mix-in class to simplify implementing a context manager

    Child classes can implement the _context_stack() function, instead of
    __aenter__() and __aexit__(). _context_stack() is called when entering the
    context. It needs to to build up an AsyncExitStack(), that is passed as an
    argument, which is then cleaned up when exiting the context.
    """

    async def _context_stack(self, stack):
        raise NotImplementedError()

    async def __aenter__(self):
        async with contextlib.AsyncExitStack() as stack:
            await self._context_stack(stack)
            self._stack = stack.pop_all()
            await self._stack.__aenter__()
            return self

    async def __aexit__(self, exc_type, exc, tb):
        await self._stack.__aexit__(exc_type, exc, tb)


@contextlib.asynccontextmanager
async def flock(f):
    """File lock as a context manager"""

    loop = asyncio.get_running_loop()
    try:
        await loop.run_in_executor(None, fcntl.flock, f.fileno(), fcntl.LOCK_EX)
        yield
    finally:
        fcntl.flock(f, fcntl.LOCK_UN)



#!/usr/bin/env python3

import asyncio
import contextlib
import contextvars
import ipaddress
import itertools
import logging
import random
import traceback

import asgineer

import not_my_board._jsonrpc as jsonrpc
import not_my_board._models as models
import not_my_board._util as util

logger = logging.getLogger(__name__)
client_ip_var = contextvars.ContextVar("client_ip")
reservation_context_var = contextvars.ContextVar("reservation_context")
valid_tokens = ("dummy-token-1", "dummy-token-2")


def hub():
    asgineer.run(asgi_app, "uvicorn", ":2092")


@asgineer.to_asgi
async def asgi_app(request):
    if isinstance(request, asgineer.WebsocketRequest):
        if request.path == "/ws-agent":
            return await _handle_agent(request)
        elif request.path == "/ws-exporter":
            return await _handle_exporter(request)
        await request.close()
        return
    elif isinstance(request, asgineer.HttpRequest):
        if request.path == "/api/v1/places":
            return await _hub.get_places()
    return 404, {}, "Page not found"


async def _handle_agent(ws):
    await _authorize_ws(ws)
    client_ip = ws.scope["client"][0]
    server = jsonrpc.Channel(ws.send, ws.receive_iter())
    await _hub.agent_communicate(client_ip, server)


async def _handle_exporter(ws):
    await _authorize_ws(ws)
    client_ip = ws.scope["client"][0]
    exporter = jsonrpc.Channel(ws.send, ws.receive_iter())
    await _hub.exporter_communicate(client_ip, exporter)


async def _authorize_ws(ws):
    try:
        auth = ws.headers["authorization"]
        scheme, token = auth.split(" ", 1)
        if scheme != "Bearer":
            raise ProtocolError(f"Invalid Authorization Scheme: {scheme}")
        if token not in valid_tokens:
            raise ProtocolError("Invalid token")
    except Exception:
        traceback.print_exc()
        await ws.close()
        return

    await ws.accept()


class Hub:
    _places = {}
    _exporters = {}
    _available = set()
    _wait_queue = []
    _reservations = {}

    def __init__(self):
        self._id_generator = itertools.count(start=1)

    @jsonrpc.hidden
    async def get_places(self):
        return {"places": [p.dict() for p in self._places.values()]}

    @jsonrpc.hidden
    async def agent_communicate(self, client_ip, rpc):
        client_ip_var.set(client_ip)
        async with self._register_agent():
            rpc.set_api_object(self)
            await rpc.communicate_forever()

    @jsonrpc.hidden
    async def exporter_communicate(self, client_ip, rpc):
        client_ip_var.set(client_ip)
        async with util.background_task(rpc.communicate_forever()) as com_task:
            export_desc = await rpc.get_place()
            with self._register_place(export_desc, rpc, client_ip):
                await com_task

    @contextlib.contextmanager
    def _register_place(self, export_desc, rpc, client_ip):
        id_ = next(self._id_generator)
        place = models.Place(id=id_, host=_unmap_ip(client_ip), **export_desc)

        try:
            logger.info("New place registered: %d", id_)
            self._places[id_] = place
            self._exporters[id_] = rpc
            self._available.add(id_)
            yield self
        finally:
            logger.info("Place disappeared: %d", id_)
            del self._places[id_]
            del self._exporters[id_]
            self._available.discard(id_)
            for candidates, _, future in self._wait_queue:
                candidates.discard(id_)
                if not candidates and not future.done():
                    future.set_exception(Exception("All candidate places are gone"))

    @contextlib.asynccontextmanager
    async def _register_agent(self):
        ctx = object()
        reservation_context_var.set(ctx)

        try:
            self._reservations[ctx] = set()
            yield
        finally:
            coros = [self.return_reservation(id_) for id_ in self._reservations[ctx]]
            results = await asyncio.gather(*coros, return_exceptions=True)
            del self._reservations[ctx]
            for result in results:
                if isinstance(result, Exception):
                    logger.warning("Error while deregistering agent: %s", result)

    async def reserve(self, candidate_ids):
        ctx = reservation_context_var.get()
        existing_candidates = {id_ for id_ in candidate_ids if id_ in self._places}
        if not existing_candidates:
            raise RuntimeError("None of the candidates exist anymore")

        available_candidates = existing_candidates & self._available
        if available_candidates:
            # TODO do something smart to get the best candidate
            reserved_id = random.choice(list(available_candidates))

            self._available.remove(reserved_id)
            self._reservations[ctx].add(reserved_id)
            logger.info("Place reserved: %d", reserved_id)
        else:
            logger.debug(
                "No places available, adding request to queue: %s",
                str(existing_candidates),
            )
            future = asyncio.get_running_loop().create_future()
            entry = (existing_candidates, ctx, future)
            self._wait_queue.append(entry)
            try:
                reserved_id = await future
            finally:
                self._wait_queue.remove(entry)

        client_ip = client_ip_var.get()
        async with util.on_error(self.return_reservation, reserved_id):
            rpc = self._exporters[reserved_id]
            await rpc.set_allowed_ips([_unmap_ip(client_ip)])

        return reserved_id

    async def return_reservation(self, place_id):
        ctx = reservation_context_var.get()
        self._reservations[ctx].remove(place_id)
        if place_id in self._places:
            for candidates, new_ctx, future in self._wait_queue:
                if place_id in candidates and not future.done():
                    self._reservations[new_ctx].add(place_id)
                    logger.info("Place returned and reserved again: %d", place_id)
                    future.set_result(place_id)
                    break
            else:
                logger.info("Place returned: %d", place_id)
                self._available.add(place_id)
                rpc = self._exporters[place_id]
                await rpc.set_allowed_ips([])
        else:
            logger.info("Place returned, but it doesn't exist: %d", place_id)


_hub = Hub()


def _unmap_ip(ip_str):
    """Resolve IPv4-mapped-on-IPv6 to an IPv4 address"""
    ip = ipaddress.ip_address(ip_str)
    unmapped = getattr(ip, "ipv4_mapped", None) or ip
    return str(unmapped)


class ProtocolError(Exception):
    pass



import asyncio

import pytest

import not_my_board._util as util

from .util import ClientVM, ExporterVM, HubVM, VMs, wait_for_ports


@pytest.fixture(scope="session")
def event_loop():
    """Redefine event_loop fixture with session scope"""

    loop = asyncio.get_event_loop_policy().new_event_loop()
    yield loop
    loop.close()


@pytest.fixture(scope="session")
async def vms():
    async with HubVM() as hub:
        await wait_for_ports(5001, 5002)
        async with ExporterVM() as exporter:
            async with ClientVM() as client:
                await util.run_concurrently(
                    hub.configure(),
                    exporter.configure(),
                    client.configure(),
                )
                await exporter.usb_attach()
                yield VMs(hub, exporter, client)



import argparse
import asyncio
import json
import logging
import pathlib
import sys

import not_my_board._client as client
import not_my_board._util as util
from not_my_board._agent import agent
from not_my_board._export import export
from not_my_board._hub import hub

try:
    from ..__about__ import __version__
except ModuleNotFoundError:
    __version__ = "dev"


# pylint: disable=too-many-statements
def main():
    parser = argparse.ArgumentParser(description="Setup, manage and use a board farm")
    parser.add_argument(
        "-V", "--version", action="version", version=f"%(prog)s {__version__}"
    )
    subparsers = parser.add_subparsers(
        dest="command",
        required=True,
        metavar="command",
        help="for more info use: %(prog)s <command> -h",
    )

    def add_subcommand(name, *args, **kwargs):
        subparser = subparsers.add_parser(name, *args, **kwargs)
        subparser.set_defaults(func=globals()[f"_{name}_command"])
        return subparser

    def add_verbose_arg(subparser):
        subparser.add_argument(
            "-v", "--verbose", action="store_true", help="Enable debug logs"
        )

    def add_cacert_arg(subparser):
        subparser.add_argument(
            "--cacert",
            type=pathlib.Path,
            action="append",
            help="path to CA certificates, which should be trusted",
        )

    subparser = add_subcommand("hub", help="start the board farm hub")
    subparser.set_defaults(verbose=True)

    subparser = add_subcommand(
        "export", help="make connected boards available in the board farm"
    )
    subparser.set_defaults(verbose=True)
    add_cacert_arg(subparser)
    subparser.add_argument("hub_url", help="http(s) URL of the hub")
    subparser.add_argument(
        "export_description", type=pathlib.Path, help="path to a export description"
    )

    subparser = add_subcommand("agent", help="start an agent")
    subparser.set_defaults(verbose=True)
    add_cacert_arg(subparser)
    subparser.add_argument("hub_url", help="http(s) URL of the hub")

    subparser = add_subcommand("reserve", help="reserve a place")
    add_verbose_arg(subparser)
    subparser.add_argument("-n", "--with-name", help="reserve under a different name")
    subparser.add_argument(
        "import_description", help="name or full path of an import description"
    )

    subparser = add_subcommand("return", help="return a place")
    add_verbose_arg(subparser)
    subparser.add_argument("name", help="name of the place to return")

    subparser = add_subcommand("attach", help="attach a reserved place")
    add_verbose_arg(subparser)
    subparser.add_argument(
        "-k",
        "--keep-others",
        action="store_true",
        help="don't return all other reservations",
    )
    subparser.add_argument(
        "name", help="name of a reserved place or an import description"
    )

    subparser = add_subcommand("detach", help="detach a reserved place")
    add_verbose_arg(subparser)
    subparser.add_argument(
        "-k", "--keep", action="store_true", help="don't return reservation"
    )
    subparser.add_argument("name", help="name of the place to attach")

    subparser = add_subcommand("list", help="list reserved places")
    add_verbose_arg(subparser)
    subparser.add_argument(
        "-n", "--no-header", action="store_true", help="Hide table header"
    )

    subparser = add_subcommand(
        "status", help="show status of attached places and its interfaces"
    )
    add_verbose_arg(subparser)
    subparser.add_argument(
        "-n", "--no-header", action="store_true", help="Hide table header"
    )

    subparser = add_subcommand("uevent", help="handle Kernel uevent for USB devices")
    add_verbose_arg(subparser)
    subparser.add_argument("devpath", help="devpath attribute of uevent")

    args = parser.parse_args()

    # Don't use escape sequences, if stdout is not a tty
    if not sys.stdout.isatty():
        for attr in dir(Format):
            if not attr.startswith("_"):
                setattr(Format, attr, "")

    if args.verbose:
        level = logging.DEBUG

        # reduce level of verbose loggers
        logging.getLogger("websockets.client").setLevel(logging.INFO)
        logging.getLogger("asyncio").setLevel(logging.INFO)
    else:
        level = logging.WARNING

    logging.basicConfig(
        format="%(asctime)s %(levelname)s: %(name)s: %(message)s", level=level
    )

    try:
        obj = args.func(args)
        if asyncio.iscoroutine(obj):
            util.run(obj, debug=True)
    except KeyboardInterrupt:
        pass


def _hub_command(_):
    hub()


async def _export_command(args):
    await export(args.hub_url, args.export_description, args.cacert)


async def _agent_command(args):
    await agent(args.hub_url, args.cacert)


async def _reserve_command(args):
    await client.reserve(args.import_description, args.with_name)


async def _return_command(args):
    await client.return_reservation(args.name)


async def _attach_command(args):
    await client.attach(args.name, args.keep_others)


async def _detach_command(args):
    await client.detach(args.name, args.keep)


async def _list_command(args):
    place_list = await client.list_()

    if not args.no_header and place_list:
        print(f"{Format.BOLD}{'Place':<16} Status{Format.RESET}")

    for entry in place_list:
        status = (
            f"{Format.GREEN}Attached"
            if entry["attached"]
            else f"{Format.YELLOW}Reserved"
        )
        print(f"{entry['place']:<16} {status}{Format.RESET}")


async def _status_command(args):
    status_list = await client.status()

    if not args.no_header and status_list:
        columns = ["Place", "Part", "Type", "Interface", "Status"]
        header = " ".join(f"{c:<16}" for c in columns).rstrip()
        print(f"{Format.BOLD}{header}{Format.RESET}")

    for entry in status_list:
        keys = ["place", "part", "type", "interface"]
        status = f"{Format.GREEN}Up" if entry["attached"] else f"{Format.RED}Down"
        row = " ".join(f"{entry[k]:<16}" for k in keys)
        row += f" {status}{Format.RESET}"
        print(row)


async def _uevent_command(args):
    await client.uevent(args.devpath)


class Format:
    RESET = "\033[0m"
    BOLD = "\033[1m"
    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"



from ._asyncio import (
    ContextStack,
    Server,
    UnixServer,
    background_task,
    cancel_tasks,
    connect,
    flock,
    on_error,
    relay_streams,
    run,
    run_concurrently,
)
from ._matching import find_matching

try:
    from tomllib import loads as toml_loads
except ModuleNotFoundError:
    from tomli import loads as toml_loads

try:
    from asyncio import timeout, timeout_at
except ImportError:
    from async_timeout import timeout, timeout_at



#!/usr/bin/env python3

import asyncio
import contextlib
import os
import pathlib
import sys

import not_my_board._jsonrpc as jsonrpc
import not_my_board._models as models
import not_my_board._util as util


async def reserve(import_description, with_name=None):
    found_import_desc = _find_import_description(import_description, with_name)

    async with agent_channel() as agent:
        await agent.reserve(found_import_desc.dict())


async def return_reservation(name):
    async with agent_channel() as agent:
        await agent.return_reservation(name)


async def attach(name, keep_others=False):
    async with agent_channel() as agent:
        reserved_names = {e["place"] for e in await agent.list()}
        if name in reserved_names:
            await agent.attach(name)

            others = reserved_names - {name}
            if not keep_others and others:
                for other in others:
                    await agent.return_reservation(name=other, force=True)
        else:
            found_import_desc = _find_import_description(name)
            await agent.reserve(found_import_desc.dict())
            await agent.attach(found_import_desc.name)

            if not keep_others and reserved_names:
                for other in reserved_names:
                    await agent.return_reservation(name=other, force=True)


async def detach(name, keep=False):
    async with agent_channel() as agent:
        await agent.detach(name)
        if not keep:
            await agent.return_reservation(name)


async def list_():
    async with agent_channel() as agent:
        return await agent.list()


async def status():
    async with agent_channel() as agent:
        return await agent.status()


async def uevent(devpath):
    # devpath has a leading "/", so joining with the / operator doesn't
    # work
    sysfs_path = pathlib.Path("/sys" + devpath)
    devname = sysfs_path.name

    pipe = pathlib.Path("/run/usbip-refresh-" + devname)
    if pipe.exists():
        with pipe.open("r+b", buffering=0) as f:
            f.write(b".")
    else:
        print(f"Loading default driver: {devname}", file=sys.stderr)
        probe_path = pathlib.Path("/sys/bus/usb/drivers_probe")
        probe_path.write_text(devname)


def _find_import_description(name, with_name=None):
    if "/" in name:
        import_description_file = pathlib.Path(name)
    else:
        path = pathlib.Path()
        home = pathlib.Path.home()

        while path != home:
            import_description_file = path / ".not-my-board" / f"{name}.toml"
            if import_description_file.is_file():
                break

            if path != path.parent:
                path = path.parent
            else:
                # we're at '/', stop loop
                path = home
        else:
            config_home = pathlib.Path(
                os.environ.get("XDG_CONFIG_HOME", home / ".config")
            )
            import_description_file = config_home / "not-my-board" / f"{name}.toml"
            if not import_description_file.is_file():
                raise ValueError(f"No import description file exists for name {name}")

    reservation_name = import_description_file.stem if not with_name else with_name
    import_description_content = util.toml_loads(import_description_file.read_text())
    import_description = models.ImportDesc(
        name=reservation_name, **import_description_content
    )

    return import_description


@contextlib.asynccontextmanager
async def agent_channel():
    socket_path = pathlib.Path("/run") / "not-my-board-agent.sock"
    reader, writer = await asyncio.open_unix_connection(socket_path)

    async def send(data):
        writer.write(data + b"\n")
        await writer.drain()

    async with jsonrpc.Channel(send, reader) as channel:
        yield channel



import pathlib


async def test_raw_usb_forwarding(vms):
    async with vms.exporter.ssh_task_root(
        "python3 -m not_my_board._usbip export 2-1", "usbip export"
    ):
        # wait for listening socket
        await vms.exporter.ssh_poll("nc -z 127.0.0.1 3240")

        async with vms.client.ssh_task_root(
            "python3 -m not_my_board._usbip import exporter.local 2-1 0",
            "usbip import",
        ):
            # wait for USB device to appear
            await vms.client.ssh_poll("test -e /dev/usbdisk")

            await vms.client.ssh("doas mount /media/usb")
            try:
                result = await vms.client.ssh("cat /media/usb/hello")
                assert result.stdout == "Hello, World!"
            finally:
                await vms.client.ssh("doas umount /media/usb")

    await vms.client.ssh_poll("! test -e /sys/bus/usb/devices/2-1")


async def test_usb_forwarding(vms):
    async with vms.hub.ssh_task("not-my-board hub", "hub"):
        # wait for listening socket
        await vms.hub.ssh_poll("nc -z 127.0.0.1 2092")

        async with vms.exporter.ssh_task_root(
            "not-my-board export http://hub.local:2092 ./src/tests/qemu-usb-place.toml",
            "export",
        ):
            await vms.client.ssh("""'doas rm -f "/run/not-my-board-agent.sock"'""")
            async with vms.client.ssh_task_root(
                "not-my-board agent http://hub.local:2092", "agent"
            ):
                # wait until exported place is registered
                await vms.client.ssh_poll(
                    "wget -q -O - http://192.168.200.1:2092/api/v1/places | grep -q qemu-usb"
                )
                # wait until agent is ready
                await vms.client.ssh_poll(
                    """'test -e "/run/not-my-board-agent.sock"'"""
                )

                await vms.client.ssh("not-my-board attach ./src/tests/qemu-usb.toml")
                # TODO attach still returns before the device is available.
                # would be nice if it blocks until the device is ready.
                await vms.client.ssh_poll("test -e /sys/bus/usb/devices/2-1")

                result = await vms.client.ssh("not-my-board status")
                status_str = result.stdout.rstrip()
                status_lines = status_str.split("\n")
                assert len(status_lines) == 2
                header = status_lines[0].split()
                status_line = status_lines[1].split()

                assert header == ["Place", "Part", "Type", "Interface", "Status"]
                assert status_line == ["qemu-usb", "flash-drive", "USB", "usb0", "Up"]

                try:
                    await vms.exporter.usb_detach()
                    await vms.client.ssh_poll("! test -e /sys/bus/usb/devices/2-1")
                finally:
                    await vms.exporter.usb_attach()

                await vms.client.ssh_poll("test -e /sys/bus/usb/devices/2-1")
                await vms.client.ssh("not-my-board detach qemu-usb")
                await vms.client.ssh("! test -e /sys/bus/usb/devices/2-1")

    # When the exporter is killed, then it should clean up and restore the
    # default USB driver.
    result = await vms.exporter.ssh("readlink /sys/bus/usb/devices/2-1/driver")
    driver_name = pathlib.Path(result.stdout).name
    assert driver_name == "usb"



import asyncio
import contextlib

import pytest

import not_my_board._hub as hubmodule
import not_my_board._jsonrpc as jsonrpc
import not_my_board._util as util

DEFAULT_EXPORTER_IP = "3.1.1.1"
DEFAULT_AGENT_IP = "6.1.1.1"


@pytest.fixture(scope="function")
def hub():
    yield hubmodule.Hub()


async def test_no_places_on_startup(hub):
    places = await hub.get_places()
    assert places["places"] == []


class FakeExporter:
    def __init__(self, register_event):
        self._register_event = register_event

    async def communicate_forever(self):
        # wait forever
        await asyncio.Event().wait()

    async def get_place(self):
        self._register_event.set()
        return {
            "port": 1234,
            "parts": [
                {
                    "compatible": [
                        "test-board",
                    ],
                    "tcp": {
                        "test-if": {
                            "host": "localhost",
                            "port": 8080,
                        },
                    },
                },
            ],
        }

    async def set_allowed_ips(self, ips):
        self._allowed_ips = ips

    @property
    def allowed_ips(self):
        return self._allowed_ips


# pylint: disable=redefined-outer-name
@contextlib.asynccontextmanager
async def register_exporter(hub, ip=DEFAULT_EXPORTER_IP):
    register_event = asyncio.Event()
    fake_exporter = FakeExporter(register_event)
    coro = hub.exporter_communicate(ip, fake_exporter)
    async with util.background_task(coro) as exporter_task:
        async with util.timeout(2):
            await register_event.wait()
        yield fake_exporter, exporter_task


async def test_register_exporter(hub):
    async with register_exporter(hub):
        places = await hub.get_places()
        assert len(places["places"]) == 1

    places = await hub.get_places()
    assert len(places["places"]) == 0


def fake_rpc_pair():
    rpc1_to_rpc2 = asyncio.Queue()
    rpc2_to_rpc1 = asyncio.Queue()

    async def receive_iter(queue):
        while True:
            data = await queue.get()
            yield data
            queue.task_done()

    rpc1 = jsonrpc.Channel(rpc1_to_rpc2.put, receive_iter(rpc2_to_rpc1))
    rpc2 = jsonrpc.Channel(rpc2_to_rpc1.put, receive_iter(rpc1_to_rpc2))
    return rpc1, rpc2


# pylint: disable=redefined-outer-name
@contextlib.asynccontextmanager
async def register_agent(hub, ip=DEFAULT_AGENT_IP):
    rpc1, rpc2 = fake_rpc_pair()
    coro = hub.agent_communicate(ip, rpc2)
    async with util.background_task(coro):
        async with util.background_task(rpc1.communicate_forever()):
            yield rpc1


async def test_reserve_place(hub):
    async with register_exporter(hub) as (exporter, _):
        async with register_agent(hub) as agent:
            places = await hub.get_places()
            candidate_ids = [places["places"][0]["id"]]
            reserved_id = await agent.reserve(candidate_ids)
            assert reserved_id == candidate_ids[0]
            assert exporter.allowed_ips == [DEFAULT_AGENT_IP]


async def test_reserve_non_existent(hub):
    async with register_agent(hub) as agent:
        candidate_ids = [42]
        with pytest.raises(jsonrpc.RemoteError) as execinfo:
            await agent.reserve(candidate_ids)
        assert "None of the candidates exist anymore" in str(execinfo.value)


async def test_reserve_queue(hub):
    async with register_exporter(hub):
        async with register_agent(hub) as agent:
            places = await hub.get_places()
            candidate_ids = [places["places"][0]["id"]]
            reserved_id = await agent.reserve(candidate_ids)

            # try to reserve same place again
            coro = agent.reserve(candidate_ids)
            async with util.background_task(coro) as reserve_task:
                await asyncio.sleep(0.001)
                # request should be in queue now
                assert not reserve_task.done()

                # when the first reservation is returned ...
                await agent.return_reservation(reserved_id)
                # ... then the second one can be fulfilled
                assert await reserve_task == reserved_id


async def test_all_places_disappear_while_trying_to_reserve(hub):
    async with register_exporter(hub) as (_, exporter_task):
        async with register_agent(hub) as agent:
            places = await hub.get_places()
            candidate_ids = [places["places"][0]["id"]]
            await agent.reserve(candidate_ids)

            # try to reserve same place again
            coro = agent.reserve(candidate_ids)
            async with util.background_task(coro) as reserve_task:
                await asyncio.sleep(0.001)
                # request should be in queue now

                # when the exporter disappears ...
                await util.cancel_tasks([exporter_task])
                # ... then the queued reservation is canceled
                with pytest.raises(Exception) as execinfo:
                    await reserve_task
                assert "All candidate places are gone" in str(execinfo.value)


async def test_one_place_disappears_while_trying_to_reserve(hub):
    async with register_exporter(hub):
        async with register_exporter(hub) as (_, exporter_task):
            async with register_agent(hub) as agent:
                places = await hub.get_places()
                # reserve both places
                candidate_ids = [p["id"] for p in places["places"]]
                await agent.reserve(candidate_ids)
                await agent.reserve(candidate_ids)

                # try to reserve both places again
                coro = agent.reserve(candidate_ids)
                async with util.background_task(coro) as reserve_task:
                    await asyncio.sleep(0.001)
                    # request should be in queue now
                    assert not reserve_task.done()

                    # when one exporter disappears ...
                    await util.cancel_tasks([exporter_task])
                    # ... then the queued reservation is still active
                    assert not reserve_task.done()


async def test_return_non_candidate(hub):
    async with register_exporter(hub):
        async with register_exporter(hub):
            async with register_agent(hub) as agent:
                places = await hub.get_places()
                # reserve both places
                candidate_ids = [p["id"] for p in places["places"]]
                await agent.reserve(candidate_ids)
                await agent.reserve(candidate_ids)

                # try to reserve place #1 again
                coro = agent.reserve(candidate_ids[:1])
                async with util.background_task(coro) as reserve_task:
                    await asyncio.sleep(0.001)
                    # request should be in queue now

                    # when place #2 is returned ...
                    await agent.return_reservation(candidate_ids[1])
                    # ... then the queued reservation is still active
                    assert not reserve_task.done()


async def test_mapped_ip_exporter(hub):
    async with register_exporter(hub, ip="::FFFF:10.0.0.8"):
        places = await hub.get_places()
        assert places["places"][0]["host"] == "10.0.0.8"


async def test_mapped_ip_agent(hub):
    async with register_exporter(hub) as (exporter, _):
        async with register_agent(hub, ip="::FFFF:10.0.0.9") as agent:
            places = await hub.get_places()
            candidate_ids = [places["places"][0]["id"]]
            await agent.reserve(candidate_ids)
            assert exporter.allowed_ips == ["10.0.0.9"]



import not_my_board._util as util

from ._protocol import Channel


class WebsocketChannel(Channel, util.ContextStack):
    def __init__(self, url, http_client, start=True, auth=None, api_obj=None):
        self._url = url
        self._http = http_client
        self._ws = None
        self._start = start
        self._auth = auth

        super().__init__(self._ws_send, self._ws_receive_iter(), api_obj)

    async def _context_stack(self, stack):
        ws = self._http.websocket(self._url, self._auth)
        self._ws = await stack.enter_async_context(ws)

        if self._start:
            await super()._context_stack(stack)

    async def _ws_receive_iter(self):
        async for message in self._ws.receive_iter():
            yield message

    async def _ws_send(self, data):
        await self._ws.send(data)



import asyncio
import collections
import contextlib
import io
import json

import pytest

import not_my_board._jsonrpc as jsonrpc
import not_my_board._util as util


class FakeApi:
    def __init__(self):
        self.event = asyncio.Event()

    async def get_42(self):
        return 42

    async def set_event(self):
        self.event.set()

    async def wait_for_event(self):
        await self.event.wait()

    async def _hidden(self):
        return "secret"

    async def error(self):
        raise RuntimeError("fake error")

    async def say_hello(self, name):
        return f"Hello, {name}!"

    @jsonrpc.hidden
    async def hidden(self):
        return "secret"


class FakeApi2:
    async def new_func(self):
        return "hi"


class FakeTransport:
    def __init__(self):
        self.receive_queue = asyncio.Queue()
        self.send_queue = asyncio.Queue()

    async def send_to_test(self, data):
        await self.receive_queue.put(data)

    async def receive_from_test(self):
        while True:
            data = await self.send_queue.get()
            yield data
            self.send_queue.task_done()

    async def send_to_jsonrpc(self, **data):
        data["jsonrpc"] = "2.0"
        raw = json.dumps(data).encode()
        await self.send_queue.put(raw)

    async def receive_from_jsonrpc(self):
        raw = await self.receive_queue.get()
        self.receive_queue.task_done()
        return json.loads(raw)

    def is_empty(self):
        return self.receive_queue.empty() and self.send_queue.empty()


Fakes = collections.namedtuple("Fakes", ["channel", "api", "transport"])


@pytest.fixture(scope="function")
async def fakes():
    transport = FakeTransport()
    api = FakeApi()
    channel = jsonrpc.Channel(
        transport.send_to_test, transport.receive_from_test(), api
    )
    async with channel:
        yield Fakes(channel, api, transport)


async def test_simple_call_execution(fakes):
    # send request
    await fakes.transport.send_to_jsonrpc(id=84, method="get_42")
    # check sent response
    message = await fakes.transport.receive_from_jsonrpc()
    assert message == {
        "jsonrpc": "2.0",
        "id": 84,
        "result": 42,
    }

    assert fakes.transport.is_empty()


async def test_execution_with_args(fakes):
    # send request
    await fakes.transport.send_to_jsonrpc(id=84, method="say_hello", params=["World"])
    # check sent response
    message = await fakes.transport.receive_from_jsonrpc()
    assert message["result"] == "Hello, World!"


async def test_execution_with_kwargs(fakes):
    # send request
    await fakes.transport.send_to_jsonrpc(
        id=84, method="say_hello", params={"name": "World"}
    )
    # check sent response
    message = await fakes.transport.receive_from_jsonrpc()
    assert message["result"] == "Hello, World!"


async def test_notification_execution(fakes):
    await fakes.transport.send_to_jsonrpc(method="set_event")

    # check if event was set
    async with util.timeout(1):
        await fakes.api.event.wait()

    # there should be no response
    assert fakes.transport.is_empty()


async def test_cancel_execution(fakes):
    await fakes.transport.send_to_jsonrpc(id=84, method="wait_for_event")

    # send cancellation request
    await fakes.transport.send_to_jsonrpc(id=85, method="rpc.cancel", params=[84])
    # check sent response
    message = await fakes.transport.receive_from_jsonrpc()
    assert message == {
        "jsonrpc": "2.0",
        "id": 85,
        "result": None,
    }

    assert fakes.transport.is_empty()


async def test_cancel_execution_invalid_id(fakes):
    # send cancellation request
    await fakes.transport.send_to_jsonrpc(id=85, method="rpc.cancel", params=[115])
    # response should still be OK
    message = await fakes.transport.receive_from_jsonrpc()
    assert message["result"] is None

    assert fakes.transport.is_empty()


async def test_prevent_hidden_function_execution(fakes):
    await fakes.transport.send_to_jsonrpc(id=84, method="_hidden")

    # call should fail
    message = await fakes.transport.receive_from_jsonrpc()
    assert message["error"]["message"] == "Method not found"


async def test_prevent_marked_hidden_function_execution(fakes):
    await fakes.transport.send_to_jsonrpc(id=84, method="hidden")

    # call should fail
    message = await fakes.transport.receive_from_jsonrpc()
    assert message["error"]["message"] == "Method not found"


async def test_return_exception(fakes):
    await fakes.transport.send_to_jsonrpc(id=84, method="error")
    # response should contain fake exception
    message = await fakes.transport.receive_from_jsonrpc()
    assert message["error"]["message"] == "fake error"


async def test_return_parse_error(fakes):
    await fakes.transport.send_to_jsonrpc(id=84, method=43)
    # response should contain parse error
    message = await fakes.transport.receive_from_jsonrpc()
    assert message["error"]["message"] == "Invalid Request"


async def test_replace_api_object(fakes):
    fakes.channel.set_api_object(FakeApi2())
    await fakes.transport.send_to_jsonrpc(id=84, method="new_func")
    # check sent response
    message = await fakes.transport.receive_from_jsonrpc()
    assert message["result"] == "hi"


async def test_simple_remote_call(fakes):
    async with util.background_task(fakes.channel.some_func()) as call:
        # check sent request
        message = await fakes.transport.receive_from_jsonrpc()
        assert message.keys() == {"jsonrpc", "method", "params", "id"}
        assert message["jsonrpc"] == "2.0"
        assert message["method"] == "some_func"
        assert not message["params"]

        # send response
        await fakes.transport.send_to_jsonrpc(id=message["id"], result=42)
        result = await call
        assert result == 42

        assert fakes.transport.is_empty()


async def test_remote_call_with_kwargs(fakes):
    async with util.background_task(fakes.channel.some_func(foo="bar")) as call:
        # check sent request
        message = await fakes.transport.receive_from_jsonrpc()
        assert message["params"] == {"foo": "bar"}

        # send response
        await fakes.transport.send_to_jsonrpc(id=message["id"], result=42)
        await call


async def test_send_notification(fakes):
    async with util.background_task(
        fakes.channel.some_notification(_notification=True)
    ) as call:
        # check sent notification
        message = await fakes.transport.receive_from_jsonrpc()
        assert message["method"] == "some_notification"
        assert "id" not in message
        await call


async def test_receive_error(fakes):
    async with util.background_task(fakes.channel.some_func()) as call:
        message = await fakes.transport.receive_from_jsonrpc()

        # send error response
        await fakes.transport.send_to_jsonrpc(
            id=message["id"],
            error={
                "code": 123,
                "message": "fake error",
                "data": {
                    "traceback": (
                        "Traceback:\n"
                        '  File "fake", line 207, in fake_func\n'
                        "    raise FakeError()\n"
                        "FakeError: fake error\n"
                    ),
                },
            },
        )

        # call should raise RemoteError
        with pytest.raises(jsonrpc.RemoteError) as execinfo:
            await call
        assert "fake error" in str(execinfo.value)


async def test_send_cancellation(fakes):
    async with util.background_task(fakes.channel.some_func()) as call:
        # wait for request to be sent
        message1 = await fakes.transport.receive_from_jsonrpc()

        # cancel the request
        call.cancel()

        # expect cancellation message
        message2 = await fakes.transport.receive_from_jsonrpc()
        assert message2["method"] == "rpc.cancel"
        assert message2["params"] == [message1["id"]]

        # send cancellation response
        await fakes.transport.send_to_jsonrpc(id=message2["id"], result=None)

        with pytest.raises(asyncio.CancelledError):
            await call

        assert fakes.transport.is_empty()


async def test_cancel_cancellation(fakes):
    async with util.background_task(fakes.channel.some_func()) as call:
        # wait for request to be sent
        await fakes.transport.receive_from_jsonrpc()

        # cancel the request
        call.cancel()

        # expect cancellation message
        await fakes.transport.receive_from_jsonrpc()

        # cancel the cancellation
        call.cancel()

        with pytest.raises(asyncio.CancelledError):
            await call

        # there should be no second cancellation message
        assert fakes.transport.is_empty()


async def test_either_args_or_kwargs(fakes):
    with pytest.raises(RuntimeError) as execinfo:
        await fakes.channel.some_func(1, foo="bar")
    assert "Use either args or kwargs" in str(execinfo.value)


async def test_prevent_hidden_function_call(fakes):
    with pytest.raises(AttributeError) as execinfo:
        # pylint: disable=protected-access
        await fakes.channel._hidden()
    assert "invalid attribute" in str(execinfo.value)


async def test_fail_call_on_parse_error(fakes):
    async with util.background_task(fakes.channel.some_func()) as call:
        # check sent request
        message = await fakes.transport.receive_from_jsonrpc()

        # send invalid error response
        await fakes.transport.send_to_jsonrpc(
            id=message["id"], error={"code": "not int"}
        )

        # call with matching ID should still fail
        with pytest.raises(jsonrpc.ProtocolError) as execinfo:
            await call
        assert '"error.code" must be an integer' in str(execinfo.value)


@contextlib.asynccontextmanager
async def capture_jsonrpc_error(fakes_):
    with contextlib.redirect_stderr(io.StringIO()) as f:
        yield f
        # send another request, to synchronize the handling of the first one
        await fakes_.transport.send_to_jsonrpc(method="set_event")
        await fakes_.api.event.wait()


async def test_parse_error_id(fakes):
    async with capture_jsonrpc_error(fakes) as f:
        await fakes.transport.send_to_jsonrpc(id=[])
    assert '"id" must be a string or number' in f.getvalue()


async def test_parse_error_method(fakes):
    async with capture_jsonrpc_error(fakes) as f:
        await fakes.transport.send_to_jsonrpc(method=33)
    assert '"method" must be a string' in f.getvalue()


async def test_parse_error_params(fakes):
    async with capture_jsonrpc_error(fakes) as f:
        await fakes.transport.send_to_jsonrpc(method="foo", params=33)
    assert '"params" must be a structured value' in f.getvalue()


async def test_parse_error_missing_id(fakes):
    async with capture_jsonrpc_error(fakes) as f:
        await fakes.transport.send_to_jsonrpc(result=None)
    assert '"id" is required' in f.getvalue()


async def test_parse_error_code(fakes):
    async with capture_jsonrpc_error(fakes) as f:
        await fakes.transport.send_to_jsonrpc(id=5, error={"code": "xxx"})
    assert '"error.code" must be an integer' in f.getvalue()


async def test_parse_error_message(fakes):
    async with capture_jsonrpc_error(fakes) as f:
        await fakes.transport.send_to_jsonrpc(id=5, error={"code": 4, "message": 3})
    assert '"error.message" must be a string' in f.getvalue()


async def test_log_error_for_notifications(fakes):
    async with capture_jsonrpc_error(fakes) as f:
        # send notification
        await fakes.transport.send_to_jsonrpc(method="error")
    assert "fake error" in f.getvalue()


async def test_pending_call_is_canceled_on_shutdown():
    transport = FakeTransport()
    channel = jsonrpc.Channel(transport.send_to_test, transport.receive_from_test())
    async with util.background_task(channel.communicate_forever()) as com_task:
        async with util.background_task(channel.some_func()) as call:
            # wait for sent message
            await transport.receive_from_jsonrpc()
            await util.cancel_tasks([com_task])

            with pytest.raises(RuntimeError) as execinfo:
                await call
            assert "Connection closed" in str(execinfo.value)


async def test_log_error_response_with_unknown_id(fakes):
    error_msg = "error for unknown call"
    async with capture_jsonrpc_error(fakes) as f:
        # send notification
        await fakes.transport.send_to_jsonrpc(
            id=43, error={"code": 2, "message": error_msg}
        )
    assert error_msg in f.getvalue()


async def test_prevent_call_if_communication_task_exited():
    transport = FakeTransport()
    api = FakeApi()
    channel = jsonrpc.Channel(
        transport.send_to_test, transport.receive_from_test(), api
    )
    async with util.background_task(channel.communicate_forever()) as com_task:
        # send one request, to synchronize com_task startup
        await transport.send_to_jsonrpc(method="set_event")
        await api.event.wait()

        # cancel communicate_forever()
        await util.cancel_tasks([com_task])

        with pytest.raises(RuntimeError) as execinfo:
            await channel.some_func()
        assert "Channel communication is already closed" in str(execinfo.value)



import asyncio
import contextlib

import pytest

import not_my_board._agent as agentmodule
import not_my_board._models as models
import not_my_board._util as util

HUB_URL = "http://fake.farm"
PLACE_1 = models.Place(
    id=1289,
    host="3.1.1.1",
    port=2000,
    parts=[
        models.ExportedPart(
            compatible=["fake-board"],
            usb={
                "usb0": models.UsbExportDesc(usbid="1-3"),
            },
            tcp={
                "ssh": models.TcpExportDesc(host="10.0.0.5", port=22),
            },
        )
    ],
)

PLACE_LOCALHOST = models.Place(
    id=338,
    host="127.0.0.1",
    port=2000,
    parts=[
        models.ExportedPart(
            compatible=["fake-board"],
            usb={
                "usb0": models.UsbExportDesc(usbid="1-3"),
            },
            tcp={
                "ssh": models.TcpExportDesc(host="10.0.0.5", port=22),
            },
        )
    ],
)

PLACE_COMPLEX = models.Place(
    id=4311,
    host="3.1.1.1",
    port=2001,
    parts=[
        models.ExportedPart(
            compatible=["fake-board"],
            usb={
                "usb0": models.UsbExportDesc(usbid="1-3"),
            },
            tcp={
                "ssh": models.TcpExportDesc(host="10.0.0.5", port=22),
            },
        ),
        models.ExportedPart(
            compatible=["fake-board"],
            tcp={
                "ssh": models.TcpExportDesc(host="10.0.0.5", port=22),
            },
        ),
    ],
)

IMPORT_DESC_1 = models.ImportDesc(
    name="fake",
    parts={
        "fake-board": models.ImportedPart(
            compatible=["fake-board"],
            usb={
                "usb0": models.UsbImportDesc(port_num=1),
            },
            tcp={
                "ssh": models.TcpImportDesc(local_port=2222),
            },
        )
    },
)

IMPORT_DESC_NOT_FOUND = models.ImportDesc(
    name="fake",
    parts={
        "fake-board": models.ImportedPart(
            compatible=["does-not-exist"],
        )
    },
)

IMPORT_DESC_COMPLEX = models.ImportDesc(
    name="fake",
    parts={
        "fake-board-ssh": models.ImportedPart(
            compatible=["fake-board"],
            tcp={
                "ssh": models.TcpImportDesc(local_port=2222),
            },
        ),
        "fake-board-usb": models.ImportedPart(
            compatible=["fake-board"],
            usb={
                "usb0": models.UsbImportDesc(port_num=1),
            },
        ),
    },
)

IMPORT_DESC_COMPLEX_NOT_FOUND = models.ImportDesc(
    name="fake",
    parts={
        "fake-board-ssh": models.ImportedPart(
            compatible=["fake-board"],
            tcp={
                "ssh": models.TcpImportDesc(local_port=2222),
            },
        ),
        "fake-board-usb": models.ImportedPart(
            compatible=["fake-board"],
            usb={
                "usb0": models.UsbImportDesc(port_num=1),
            },
        ),
        "fake-board-any": models.ImportedPart(
            compatible=["fake-board"],
        ),
    },
)


class FakeHub:
    def __init__(self):
        self.reserve_request = None
        self.reserved = set()
        self.reserve_continue = asyncio.Event()
        self.reserve_pending = asyncio.Event()
        self.reserve_continue.set()

    async def reserve(self, candidate_ids):
        self.reserve_pending.set()
        try:
            await self.reserve_continue.wait()
        finally:
            self.reserve_pending.clear()

        self.reserve_request = candidate_ids
        self.reserved.add(candidate_ids[0])
        return candidate_ids[0]

    async def return_reservation(self, place_id):
        self.reserved.remove(place_id)


class NopServer:
    async def serve_forever(self):
        await asyncio.Event().wait()


class FakeAgentIO:
    def __init__(self):
        self.places = []
        self.attached = {}
        self.detach_event = {}
        self.port_forwards = {}

    @contextlib.asynccontextmanager
    async def hub_rpc(self):
        self.hub = FakeHub()
        yield self.hub

    @contextlib.asynccontextmanager
    async def unix_server(self, api_obj):
        self.agent_api = api_obj
        yield NopServer()

    async def get_places(self):
        return self.places

    @staticmethod
    async def usbip_refresh_status():
        pass

    def usbip_is_attached(self, vhci_port):
        return vhci_port in self.attached

    async def usbip_attach(self, proxy, target, port_num, usbid):
        if port_num in self.detach_event:
            await self.detach_event[port_num].wait()
        self.attached[port_num] = (proxy, target, usbid)
        self.detach_event[port_num] = asyncio.Event()
        return port_num

    def usbip_detach(self, vhci_port):
        if vhci_port in self.attached:
            del self.attached[vhci_port]
            self.detach_event[vhci_port].set()
            del self.detach_event[vhci_port]

    async def port_forward(self, ready_event, proxy, target, local_port):
        self.port_forwards[local_port] = (proxy, target)
        ready_event.set()
        try:
            await asyncio.Event().wait()
        finally:
            del self.port_forwards[local_port]


@pytest.fixture(scope="function")
async def agent_io():
    io = FakeAgentIO()
    async with agentmodule.Agent(HUB_URL, io) as agent:
        async with util.background_task(agent.serve_forever()):
            yield io


async def test_idle_list(agent_io):
    list_ = await agent_io.agent_api.list()
    assert list_ == []


async def test_idle_status(agent_io):
    status = await agent_io.agent_api.status()
    assert status == []


async def test_reserve(agent_io):
    agent_io.places = [PLACE_1]
    await agent_io.agent_api.reserve(IMPORT_DESC_1.dict())
    assert agent_io.hub.reserve_request == [PLACE_1.id]
    assert agent_io.hub.reserved == {PLACE_1.id}


async def test_attach(agent_io):
    agent_io.places = [PLACE_1]

    await agent_io.agent_api.reserve(IMPORT_DESC_1.dict())
    await agent_io.agent_api.attach(IMPORT_DESC_1.name)
    port_num = IMPORT_DESC_1.parts["fake-board"].usb["usb0"].port_num
    proxy = (PLACE_1.host, PLACE_1.port)
    usbip_target = ("usb.not-my-board.localhost", 3240)
    usbip = PLACE_1.parts[0].usb["usb0"].usbid
    assert agent_io.attached == {port_num: (proxy, usbip_target, usbip)}
    local_port = IMPORT_DESC_1.parts["fake-board"].tcp["ssh"].local_port
    tcp_target = tuple(
        getattr(PLACE_1.parts[0].tcp["ssh"], k) for k in ("host", "port")
    )
    assert agent_io.port_forwards == {local_port: (proxy, tcp_target)}


async def test_list_place_1(agent_io):
    agent_io.places = [PLACE_1]
    await agent_io.agent_api.reserve(IMPORT_DESC_1.dict())
    list_ = await agent_io.agent_api.list()
    assert list_ == [{"place": "fake", "attached": False}]


async def test_list_place_1_attached(agent_io):
    agent_io.places = [PLACE_1]
    await agent_io.agent_api.reserve(IMPORT_DESC_1.dict())
    await agent_io.agent_api.attach(IMPORT_DESC_1.name)
    list_ = await agent_io.agent_api.list()
    assert list_ == [{"place": "fake", "attached": True}]


async def test_status_place_1(agent_io):
    agent_io.places = [PLACE_1]
    await agent_io.agent_api.reserve(IMPORT_DESC_1.dict())
    status = await agent_io.agent_api.status()
    assert len(status) == 2
    usb0_status = {
        "place": "fake",
        "part": "fake-board",
        "interface": "usb0",
        "type": "USB",
        "attached": False,
    }
    ssh_status = {
        "place": "fake",
        "part": "fake-board",
        "interface": "ssh",
        "type": "TCP",
        "attached": False,
    }
    assert usb0_status in status
    assert ssh_status in status


async def test_status_place_1_attached(agent_io):
    agent_io.places = [PLACE_1]
    await agent_io.agent_api.reserve(IMPORT_DESC_1.dict())
    await agent_io.agent_api.attach(IMPORT_DESC_1.name)
    status = await agent_io.agent_api.status()
    assert len(status) == 2
    assert status[0]["attached"] is True
    assert status[1]["attached"] is True


async def test_reserve_twice(agent_io):
    agent_io.places = [PLACE_1]
    await agent_io.agent_api.reserve(IMPORT_DESC_1.dict())
    with pytest.raises(RuntimeError) as execinfo:
        await agent_io.agent_api.reserve(IMPORT_DESC_1.dict())
    assert "is already reserved" in str(execinfo.value)


async def test_return_reservation(agent_io):
    agent_io.places = [PLACE_1]
    await agent_io.agent_api.reserve(IMPORT_DESC_1.dict())
    await agent_io.agent_api.return_reservation(IMPORT_DESC_1.name)
    assert not agent_io.hub.reserved


async def test_detach(agent_io):
    agent_io.places = [PLACE_1]
    await agent_io.agent_api.reserve(IMPORT_DESC_1.dict())
    await agent_io.agent_api.attach(IMPORT_DESC_1.name)
    await agent_io.agent_api.detach(IMPORT_DESC_1.name)
    assert not agent_io.attached
    assert not agent_io.port_forwards

    status = await agent_io.agent_api.status()
    assert len(status) == 2
    assert not status[0]["attached"]
    assert not status[1]["attached"]


async def test_return_reservation_while_attached(agent_io):
    agent_io.places = [PLACE_1]
    await agent_io.agent_api.reserve(IMPORT_DESC_1.dict())
    await agent_io.agent_api.attach(IMPORT_DESC_1.name)
    with pytest.raises(RuntimeError) as execinfo:
        await agent_io.agent_api.return_reservation(IMPORT_DESC_1.name)
    assert "is still attached" in str(execinfo.value)


async def test_force_return_reservation(agent_io):
    agent_io.places = [PLACE_1]
    await agent_io.agent_api.reserve(IMPORT_DESC_1.dict())
    await agent_io.agent_api.attach(IMPORT_DESC_1.name)
    await agent_io.agent_api.return_reservation(name=IMPORT_DESC_1.name, force=True)
    assert not agent_io.attached
    assert not agent_io.port_forwards
    assert not agent_io.hub.reserved


async def test_return_unreserved_place(agent_io):
    with pytest.raises(RuntimeError) as execinfo:
        await agent_io.agent_api.return_reservation(IMPORT_DESC_1.name)
    assert "is not reserved" in str(execinfo.value)


async def test_no_match_found(agent_io):
    agent_io.places = [PLACE_1]
    with pytest.raises(RuntimeError) as execinfo:
        await agent_io.agent_api.reserve(IMPORT_DESC_NOT_FOUND.dict())
    assert "No matching place found" in str(execinfo.value)


async def test_localhost_exporter(agent_io):
    """This happens if the Hub and an Exporter run on the same host"""
    agent_io.places = [PLACE_LOCALHOST]

    await agent_io.agent_api.reserve(IMPORT_DESC_1.dict())
    await agent_io.agent_api.attach(IMPORT_DESC_1.name)
    port_num = IMPORT_DESC_1.parts["fake-board"].usb["usb0"].port_num
    # place host needs to be replaced with hub host
    proxy = ("fake.farm", PLACE_1.port)
    assert agent_io.attached[port_num][0] == proxy


async def test_attach_twice(agent_io):
    agent_io.places = [PLACE_1]
    await agent_io.agent_api.reserve(IMPORT_DESC_1.dict())
    await agent_io.agent_api.attach(IMPORT_DESC_1.name)
    with pytest.raises(RuntimeError) as execinfo:
        await agent_io.agent_api.attach(IMPORT_DESC_1.name)
    assert "is already attached" in str(execinfo.value)


async def test_detach_twice(agent_io):
    agent_io.places = [PLACE_1]
    await agent_io.agent_api.reserve(IMPORT_DESC_1.dict())
    await agent_io.agent_api.attach(IMPORT_DESC_1.name)
    await agent_io.agent_api.detach(IMPORT_DESC_1.name)
    with pytest.raises(RuntimeError) as execinfo:
        await agent_io.agent_api.detach(IMPORT_DESC_1.name)
    assert "is not attached" in str(execinfo.value)


async def test_complex_match(agent_io):
    agent_io.places = [PLACE_COMPLEX]
    await agent_io.agent_api.reserve(IMPORT_DESC_COMPLEX.dict())
    assert await agent_io.agent_api.list()


async def test_complex_match_not_found(agent_io):
    agent_io.places = [PLACE_COMPLEX]
    with pytest.raises(RuntimeError) as execinfo:
        await agent_io.agent_api.reserve(IMPORT_DESC_COMPLEX_NOT_FOUND.dict())
    assert "No matching place found" in str(execinfo.value)


async def test_reserve_twice_concurrently(agent_io):
    agent_io.places = [PLACE_1]

    # make the reserve call blocking
    agent_io.hub.reserve_continue.clear()

    # start two reserve tasks in parallel
    coro_1 = agent_io.agent_api.reserve(IMPORT_DESC_1.dict())
    coro_2 = agent_io.agent_api.reserve(IMPORT_DESC_1.dict())
    async with util.background_task(coro_1) as task_1:
        async with util.background_task(coro_2) as task_2:
            # wait until one blocks
            await agent_io.hub.reserve_pending.wait()

            # the other one should block until the first one finishes
            assert not task_1.done()
            assert not task_2.done()

            # unblock reserve call
            agent_io.hub.reserve_continue.set()

            # now one should finish successfully and the other one should fail
            results = await asyncio.gather(task_1, task_2, return_exceptions=True)

            assert None in results
            exception = results[0] or results[1]
            assert "is already reserved" in str(exception)

            assert len(await agent_io.agent_api.list()) == 1



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import pathlib
import subprocess

project_dir = pathlib.Path(__file__).parents[1]

project = "not-my-board"
copyright = "2023-present, Simon Holesch"
author = "Simon Holesch"
release = subprocess.run(
    "scripts/get_version", cwd=project_dir, capture_output=True, text=True, check=True
).stdout.strip()

extensions = [
    "myst_parser",
    "sphinx_copybutton",
    "sphinxext.opengraph",
]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

html_theme = "furo"
html_title = f"{project} Documentation"

myst_enable_extensions = [
    "deflist",
    "colon_fence",
]

# exclude prompts and output from copies
copybutton_exclude = ".linenos, .gp, .go"

# add search console verification
html_extra_path = ["googlef248cffc234a230d.html"]

# add custom CSS
html_static_path = ["_static"]
html_css_files = [
    "css/custom.css",
]



#!/usr/bin/env python3

import asyncio
import contextlib
import functools
import ipaddress
import logging
import os
import pathlib
import shutil
import traceback
import urllib.parse
import weakref
from dataclasses import dataclass, field
from typing import List, Tuple

import not_my_board._http as http
import not_my_board._jsonrpc as jsonrpc
import not_my_board._models as models
import not_my_board._usbip as usbip
import not_my_board._util as util

logger = logging.getLogger(__name__)
USBIP_REMOTE = ("usb.not-my-board.localhost", 3240)
Address = Tuple[str, int]


async def agent(hub_url, ca_files):
    io = _AgentIO(hub_url, http.Client(ca_files))
    async with Agent(hub_url, io) as agent_:
        await agent_.serve_forever()


class _AgentIO:
    def __init__(self, hub_url, http_client):
        self._hub_url = hub_url
        self._http = http_client

    @contextlib.asynccontextmanager
    async def hub_rpc(self):
        auth = "Bearer dummy-token-1"
        url = f"{self._hub_url}/ws-agent"
        async with jsonrpc.WebsocketChannel(url, self._http, auth=auth) as rpc:
            yield rpc

    @contextlib.asynccontextmanager
    async def unix_server(self, api_obj):
        socket_path = pathlib.Path("/run") / "not-my-board-agent.sock"

        connection_handler = functools.partial(self._handle_unix_client, api_obj)
        async with util.UnixServer(connection_handler, socket_path) as unix_server:
            os.chmod(socket_path, 0o660)
            try:
                shutil.chown(socket_path, group="not-my-board")
            except Exception as e:
                logger.warning(
                    'Failed to change group on agent socket "%s": %s', socket_path, e
                )

            yield unix_server

    @staticmethod
    async def _handle_unix_client(api_obj, reader, writer):
        async def send(data):
            writer.write(data + b"\n")
            await writer.drain()

        channel = jsonrpc.Channel(send, reader, api_obj)
        await channel.communicate_forever()

    async def get_places(self):
        response = await self._http.get_json(f"{self._hub_url}/api/v1/places")
        return [models.Place(**p) for p in response["places"]]

    @staticmethod
    async def usbip_refresh_status():
        await usbip.refresh_vhci_status()

    @staticmethod
    def usbip_is_attached(vhci_port):
        return usbip.is_attached(vhci_port)

    async def usbip_attach(self, proxy, target, port_num, usbid):
        tunnel = self._http.open_tunnel(*proxy, *target)
        async with tunnel as (reader, writer, trailing_data):
            if trailing_data:
                raise ProtocolError("USB/IP implementation cannot handle trailing data")
            return await usbip.attach(reader, writer, usbid, port_num)

    @staticmethod
    def usbip_detach(vhci_port):
        usbip.detach(vhci_port)

    async def port_forward(self, ready_event, proxy, target, local_port):
        localhost = "127.0.0.1"
        connection_handler = functools.partial(
            self._handle_port_forward_client, proxy, target
        )
        async with util.Server(connection_handler, localhost, local_port) as server:
            ready_event.set()
            await server.serve_forever()

    async def _handle_port_forward_client(self, proxy, target, client_r, client_w):
        async with self._http.open_tunnel(*proxy, *target) as (
            remote_r,
            remote_w,
            trailing_data,
        ):
            client_w.write(trailing_data)
            await client_w.drain()
            await util.relay_streams(client_r, client_w, remote_r, remote_w)


class Agent(util.ContextStack):
    def __init__(self, hub_url, io):
        url = urllib.parse.urlsplit(hub_url)
        self._hub_host = url.netloc.split(":")[0]
        self._io = io
        self._locks = weakref.WeakValueDictionary()
        self._reservations = {}

    async def _context_stack(self, stack):
        self._hub = await stack.enter_async_context(self._io.hub_rpc())
        stack.push_async_callback(self._cleanup)
        self._unix_server = await stack.enter_async_context(self._io.unix_server(self))

    async def serve_forever(self):
        await self._unix_server.serve_forever()

    async def _cleanup(self):
        coros = [t.close() for r in self._reservations.values() for t in r.tunnels]

        await util.run_concurrently(*coros)

    async def reserve(self, import_description):
        import_description = models.ImportDesc(**import_description)
        name = import_description.name

        async with self._name_lock(name):
            if name in self._reservations:
                raise RuntimeError(f'A place named "{name}" is already reserved')

            places = await self._io.get_places()

            candidates = _filter_places(import_description, places)
            if not candidates:
                raise RuntimeError("No matching place found")

            candidate_ids = list(candidates)
            place_id = await self._hub.reserve(candidate_ids)

            tunnels = [
                desc.tunnel_cls(desc, self._hub_host, self._io)
                for desc in candidates[place_id]
            ]
            self._reservations[name] = _Reservation(place_id, tunnels)

    async def return_reservation(self, name, force=False):
        async with self._reservation(name) as reservation:
            if reservation.is_attached:
                if force:
                    await self._detach_reservation(reservation)
                else:
                    raise RuntimeError(f'Place "{name}" is still attached')
            await self._hub.return_reservation(reservation.place_id)
            del self._reservations[name]

    async def attach(self, name):
        async with self._reservation(name) as reservation:
            if reservation.is_attached:
                raise RuntimeError(f'Place "{name}" is already attached')

            coros = [t.open() for t in reservation.tunnels]

            async with util.on_error(self._detach_reservation, reservation):
                await util.run_concurrently(*coros)
                reservation.is_attached = True

    async def detach(self, name):
        async with self._reservation(name) as reservation:
            if not reservation.is_attached:
                raise RuntimeError(f'Place "{name}" is not attached')

            await self._detach_reservation(reservation)

    async def _detach_reservation(self, reservation):
        coros = [t.close() for t in reservation.tunnels]
        await util.run_concurrently(*coros)
        reservation.is_attached = False

    @contextlib.asynccontextmanager
    async def _name_lock(self, name):
        if name not in self._locks:
            lock = asyncio.Lock()
            self._locks[name] = lock
        else:
            lock = self._locks[name]

        async with lock:
            yield

    @contextlib.asynccontextmanager
    async def _reservation(self, name):
        async with self._name_lock(name):
            if name not in self._reservations:
                raise RuntimeError(f'A place named "{name}" is not reserved')
            yield self._reservations[name]

    async def list(self):
        return [
            {"place": name, "attached": reservation.is_attached}
            for name, reservation in self._reservations.items()
        ]

    async def status(self):
        await self._io.usbip_refresh_status()

        return [
            {
                "place": tunnel.place_name,
                "part": tunnel.part_name,
                "interface": tunnel.iface_name,
                "type": tunnel.type_name,
                "attached": tunnel.is_attached(),
            }
            for name, reservation in self._reservations.items()
            for tunnel in reservation.tunnels
        ]


def _filter_places(import_description, places):
    candidates = {}

    imported_part_sets = [
        (name, _part_to_set(imported_part))
        for name, imported_part in import_description.parts.items()
    ]

    for place in places:
        matching = _find_matching(imported_part_sets, place)
        if matching:
            candidates[place.id] = _create_tunnel_descriptions(
                import_description, place, matching
            )

    return candidates


# pylint: disable=too-many-locals
def _create_tunnel_descriptions(import_description, place, matching):
    place_name = import_description.name
    proxy = (place.host, place.port)
    tunnel_descs = set()

    for part_name, place_part_idx in matching:
        imported_part = import_description.parts[part_name]
        place_part = place.parts[place_part_idx]

        for iface_name, usb_import_description in imported_part.usb.items():
            usbid = place_part.usb[iface_name].usbid
            port_num = usb_import_description.port_num
            tunnel_desc = _UsbTunnelDesc(
                place_name, part_name, iface_name, proxy, usbid, port_num
            )
            tunnel_descs.add(tunnel_desc)

        for iface_name, tcp_import_description in imported_part.tcp.items():
            host = place_part.tcp[iface_name].host
            port = place_part.tcp[iface_name].port
            remote = (host, port)
            local_port = tcp_import_description.local_port
            tunnel_desc = _TcpTunnelDesc(
                place_name, part_name, iface_name, proxy, remote, local_port
            )
            tunnel_descs.add(tunnel_desc)

    return tunnel_descs


def _find_matching(imported_part_sets, place):
    match_graph = {}
    for name, imported_part_set in imported_part_sets:
        match_graph[name] = []
        for place_part_idx, place_part in enumerate(place.parts):
            place_part_set = _part_to_set(place_part)
            if imported_part_set.issubset(place_part_set):
                match_graph[name].append(place_part_idx)

        if not match_graph[name]:
            return None

    matching = []
    for name, matches in match_graph.items():
        if len(matches) != 1:
            # Complex matching
            matching_dict = util.find_matching(match_graph)
            if len(match_graph) == len(matching_dict):
                return matching_dict.items()
            else:
                return None

        matching.append((name, matches[0]))

    return matching


def _part_to_set(part):
    return (
        {f"compatible:{c}" for c in part.compatible}
        | {f"usb:{k}" for k in part.usb}
        | {f"tcp:{k}" for k in part.tcp}
    )


class _Tunnel:
    _ready_timeout = 5

    def __init__(self, desc, hub_host, io):
        self._desc = desc
        self._io = io
        self._task = None
        self._ready_event = asyncio.Event()

        host = self.proxy[0]
        if ipaddress.ip_address(host).is_loopback:
            host = hub_host
        self._proxy = (host, self.proxy[1])

    def __getattr__(self, attr):
        return getattr(self._desc, attr)

    async def open(self):
        if not self._task:
            logger.debug("%s: Opening %s tunnel", self.name, self.type_name)
            self._ready_event.clear()
            self._task = asyncio.create_task(self._task_func())
            self._task.add_done_callback(self._task_done_callback)

        try:
            await asyncio.wait_for(self._ready_event.wait(), self._ready_timeout)
        except asyncio.TimeoutError:
            logger.warning("%s: Opening %s tunnel timed out", self.name, self.type_name)

    def _task_done_callback(self, _):
        self._task = None

    async def close(self):
        if self._task:
            await util.cancel_tasks([self._task])

    def is_attached(self):
        return bool(self._task and self._ready_event.is_set())


class _UsbTunnel(_Tunnel):
    _vhci_port = None

    async def close(self):
        await super().close()
        if self._vhci_port is not None:
            self._io.usbip_detach(self._vhci_port)

    async def _task_func(self):
        retry_timeout = 1
        while True:
            try:
                self._vhci_port = await self._io.usbip_attach(
                    self._proxy, USBIP_REMOTE, self.port_num, self.usbid
                )
                logger.debug("%s: USB device attached", self.name)
                self._ready_event.set()
                retry_timeout = 1
            except Exception:
                traceback.print_exc()
                await asyncio.sleep(retry_timeout)
                retry_timeout = min(2 * retry_timeout, 30)

    def is_attached(self):
        if self._vhci_port is None:
            return False
        return self._io.usbip_is_attached(self._vhci_port)


class _TcpTunnel(_Tunnel):
    async def _task_func(self):
        await self._io.port_forward(
            self._ready_event, self._proxy, self.remote, self.local_port
        )


@dataclass(frozen=True)
class _TunnelDesc:
    place_name: str
    part_name: str
    iface_name: str
    proxy: Address

    @property
    def name(self):
        return f"{self.place_name}.{self.part_name}.{self.iface_name}"


@dataclass(frozen=True)
class _UsbTunnelDesc(_TunnelDesc):
    usbid: str
    port_num: int
    type_name: str = field(default="USB", init=False)
    tunnel_cls: type = field(default=_UsbTunnel, init=False)


@dataclass(frozen=True)
class _TcpTunnelDesc(_TunnelDesc):
    remote: Address
    local_port: int
    type_name: str = field(default="TCP", init=False)
    tunnel_cls: type = field(default=_TcpTunnel, init=False)


@dataclass
class _Reservation:
    place_id: int
    is_attached: bool = field(default=False, init=False)
    tunnels: List[_Tunnel]


class ProtocolError(Exception):
    pass



import asyncio
import collections
import contextlib
import pathlib
import sys

import not_my_board._util as util

VMs = collections.namedtuple("VMs", ["hub", "exporter", "client"])
ShResult = collections.namedtuple("ShResult", ["stdout", "stderr", "returncode"])


class _VM(util.ContextStack):
    _name = ""

    async def _context_stack(self, stack):
        await stack.enter_async_context(
            sh_task(f"./scripts/vmctl run {self._name}", f"vm {self._name}")
        )

    async def configure(self):
        await sh(
            f"./scripts/vmctl configure {self._name}", prefix=f"configure {self._name}"
        )

    def ssh_task(self, cmd, *args, **kwargs):
        return sh_task(
            f"./scripts/vmctl ssh {self._name} while-stdin " + cmd,
            *args,
            terminate=False,
            **kwargs,
        )

    def ssh_task_root(self, cmd, *args, **kwargs):
        return sh_task(
            f"./scripts/vmctl ssh {self._name} doas while-stdin " + cmd,
            *args,
            terminate=False,
            **kwargs,
        )

    async def ssh(self, cmd, *args, **kwargs):
        return await sh(f"./scripts/vmctl ssh {self._name} " + cmd, *args, **kwargs)

    async def ssh_poll(self, cmd, timeout=None):
        return await sh_poll(f"./scripts/vmctl ssh {self._name} " + cmd, timeout)


class HubVM(_VM):
    _name = "hub"


class ExporterVM(_VM):
    _name = "exporter"

    async def usb_attach(self):
        await sh("./scripts/vmctl usb attach")

    async def usb_detach(self):
        await sh("./scripts/vmctl usb detach")


class ClientVM(_VM):
    _name = "client"


async def sh(cmd, check=True, strip=True, prefix=None):
    proc = await asyncio.create_subprocess_shell(
        cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
    )

    stdout, _ = await util.run_concurrently(
        proc.stdout.read(), _log_output(proc.stderr, cmd, prefix)
    )
    await proc.wait()
    if check and proc.returncode:
        raise RuntimeError(f"{cmd!r} exited with {proc.returncode}")

    stdout = stdout.decode("utf-8")
    if strip:
        stdout = stdout.rstrip()

    return ShResult(stdout, None, proc.returncode)


@contextlib.asynccontextmanager
async def sh_task(cmd, prefix=None, terminate=True):
    # need to exec, otherwise only the shell process is killed with
    # proc.terminate()
    proc = await asyncio.create_subprocess_shell(
        f"exec {cmd}",
        stdin=asyncio.subprocess.PIPE,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.STDOUT,
    )

    logging_task = None
    try:
        logging_task = asyncio.create_task(_log_output(proc.stdout, cmd, prefix))
        yield
    finally:
        proc.stdin.close()
        await proc.stdin.wait_closed()

        if terminate:
            with contextlib.suppress(ProcessLookupError):
                proc.terminate()

        await proc.wait()
        if logging_task:
            await logging_task


async def sh_poll(cmd, timeout=None):
    if pathlib.Path("/dev/kvm").exists():
        if timeout is None:
            timeout = 7
        interval = 0.1
    else:
        if timeout is None:
            timeout = 60
        interval = 1

    async def poll_loop():
        while True:
            result = await sh(cmd, check=False)
            if result.returncode == 0:
                break
            await asyncio.sleep(interval)

    await asyncio.wait_for(poll_loop(), timeout)


async def _log_output(stream, cmd, prefix):
    if prefix is None:
        prefix = f"[{cmd}] ".encode()
    else:
        prefix = f"[{prefix}] ".encode()

    async for line in stream:
        sys.stderr.buffer.write(prefix + line)
        sys.stderr.buffer.flush()


async def wait_for_ports(*ports, timeout=7):
    async with util.timeout(timeout):
        while True:
            try:
                for port in ports:
                    async with util.connect("127.0.0.1", port):
                        pass
            except ConnectionRefusedError:
                await asyncio.sleep(0.1)
                continue
            break



import pathlib

import pytest

import not_my_board._http as http

from .util import sh, sh_task, wait_for_ports

project_dir = pathlib.Path(__file__).parents[1]


@pytest.fixture(scope="session")
async def tinyproxy():
    conf = project_dir / "tests/tinyproxy.conf"
    async with sh_task(f"tinyproxy -d -c {conf}", "tinyproxy"):
        await wait_for_ports(8888)
        yield "http://127.0.0.1:8888"


async def test_proxy_connect(tinyproxy):
    async with sh_task("not-my-board hub", "hub"):
        await wait_for_ports(2092)

        client = http.Client(proxies={"http": tinyproxy})
        response = await client.get_json("http://127.0.0.1:2092/api/v1/places")
        assert response == {"places": []}


async def test_proxy_connect_https(tinyproxy):
    root_key = project_dir / "tests/.cache/not-my-board-root-ca.key"
    root_cert = project_dir / "tests/.cache/not-my-board-root-ca.crt"
    key_file = project_dir / "tests/.cache/not-my-board.key"
    cert_file = project_dir / "tests/.cache/not-my-board.crt"
    if not key_file.exists():
        key_file.parent.mkdir(exist_ok=True)
        await sh(
            "openssl req "
            "-x509 "
            "-newkey ec "
            "-pkeyopt ec_paramgen_curve:secp384r1 "
            "-days 365000 "
            "-nodes "
            f"-keyout {root_key} "
            f"-out {root_cert} "
            "-subj '/CN=not-my-board-root-ca'"
        )
        hostname = "hub.local"
        await sh(
            "openssl req "
            "-x509 "
            f"-CA {root_cert} "
            f"-CAkey {root_key} "
            "-newkey ec "
            "-pkeyopt ec_paramgen_curve:prime256v1 "
            "-days 365000 "
            "-nodes "
            f"-keyout {key_file} "
            f"-out {cert_file} "
            f"-subj '/CN={hostname}' "
            f"-addext 'subjectAltName=DNS:{hostname},DNS:*.{hostname},IP:127.0.0.1' "
            "-addext 'basicConstraints=CA:FALSE' "
            "-addext 'keyUsage=digitalSignature,keyEncipherment' "
            "-addext 'extendedKeyUsage=serverAuth'"
        )

    async with sh_task(
        (
            "uvicorn --port 2092 "
            f"--ssl-keyfile {key_file} "
            f"--ssl-certfile {cert_file} "
            "not_my_board:asgi_app"
        ),
        "hub",
    ):
        await wait_for_ports(2092)

        client = http.Client(ca_files=[root_cert], proxies={"https": tinyproxy})
        response = await client.get_json("https://127.0.0.1:2092/api/v1/places")
        assert response == {"places": []}


@pytest.mark.parametrize(
    "host,no_proxy_env,expected",
    [
        # Test cases taken from curl
        ("www.example.com", "localhost,.example.com,.example.de", True),
        ("www.example.com.", "localhost,.example.com,.example.de", True),
        ("example.com", "localhost,.example.com,.example.de", True),
        ("example.com.", "localhost,.example.com,.example.de", True),
        ("www.example.com", "localhost,.example.com.,.example.de", True),
        ("www.example.com", "localhost,www.example.com.,.example.de", True),
        ("example.com", "localhost,example.com,.example.de", True),
        ("example.com.", "localhost,example.com,.example.de", True),
        ("nexample.com", "localhost,example.com,.example.de", False),
        ("www.example.com", "localhost,example.com,.example.de", True),
        ("127.0.0.1", "127.0.0.1,localhost", True),
        ("127.0.0.1", "127.0.0.1,localhost,", True),
        ("127.0.0.1", "127.0.0.1/8,localhost,", True),
        ("127.0.0.1", "127.0.0.1/28,localhost,", True),
        ("127.0.0.1", "127.0.0.1/31,localhost,", True),
        ("127.0.0.1", "localhost,127.0.0.1", True),
        ("localhost", "localhost,127.0.0.1", True),
        ("localhost", "127.0.0.1,localhost", True),
        ("foobar", "barfoo", False),
        ("foobar", "foobar", True),
        ("192.168.0.1", "foobar", False),
        ("192.168.0.1", "192.168.0.0/16", True),
        ("192.168.0.1", "192.168.0.0/24", True),
        ("192.168.0.1", "192.168.0.0/32", False),
        ("192.168.0.1", "192.168.0.0", False),
        ("192.168.1.1", "192.168.0.0/24", False),
        ("192.168.1.1", "foo, bar, 192.168.0.0/24", False),
        ("192.168.1.1", "foo, bar, 192.168.0.0/16", True),
        ("[::1]", "foo, bar, 192.168.0.0/16", False),
        ("[::1]", "foo, bar, ::1/64", True),
        ("bar", "foo, bar, ::1/64", True),
        ("BAr", "foo, bar, ::1/64", True),
        ("BAr", "foo,,,,,              bar, ::1/64", True),
        ("www.example.com", "foo, .example.com", True),
        ("www.example.com", "www2.example.com, .example.net", False),
        ("example.com", ".example.com, .example.net", True),
        ("nonexample.com", ".example.com, .example.net", False),
        # Test cases taken from CPython without host:port cases
        ("anotherdomain.com", "localhost, anotherdomain.com", True),
        ("localhost", "localhost, anotherdomain.com, .d.o.t", True),
        ("LocalHost", "localhost, anotherdomain.com, .d.o.t", True),
        ("LOCALHOST", "localhost, anotherdomain.com, .d.o.t", True),
        (".localhost", "localhost, anotherdomain.com, .d.o.t", True),
        ("foo.d.o.t", "localhost, anotherdomain.com, .d.o.t", True),
        ("d.o.t", "localhost, anotherdomain.com, .d.o.t", True),
        ("prelocalhost", "localhost, anotherdomain.com, .d.o.t", False),
        ("newdomain.com", "localhost, anotherdomain.com, .d.o.t", False),
        ("newdomain.com", "*", True),
        ("anotherdomain.com", "*, anotherdomain.com", True),
        ("newdomain.com", "*, anotherdomain.com", False),
        ("localhost\n", "localhost, anotherdomain.com", False),
    ],
)
def test_is_proxy_disabled(host, no_proxy_env, expected):
    assert http.is_proxy_disabled(host, no_proxy_env) == expected



#!/usr/bin/python3

import asyncio
import codecs
import contextlib
import ipaddress
import json
import logging
import ssl
import urllib.parse
import urllib.request
from dataclasses import dataclass
from typing import Optional, Union

import h11
import websockets
import websockets.frames
import websockets.http11
import websockets.uri

import not_my_board._util as util

logger = logging.getLogger(__name__)
UTF8Decoder = codecs.getincrementaldecoder("utf-8")


class ProtocolError(Exception):
    pass


class Client:
    def __init__(self, ca_files=None, proxies=None):
        self._ssl_ctx = ssl.create_default_context()
        if ca_files:
            for ca_file in ca_files:
                self._ssl_ctx.load_verify_locations(cafile=ca_file)

        if proxies is None:
            proxies = urllib.request.getproxies()
        self._proxies = {
            scheme: self._parse_url(url) for scheme, url in proxies.items()
        }

    async def get_json(self, url):
        return await self._request_json("GET", url)

    async def post_form(self, url, params):
        content_type = "application/x-www-form-urlencoded"
        body = urllib.parse.urlencode(params).encode()
        return await self._request_json("POST", url, content_type, body)

    async def _request_json(self, method, url, content_type=None, body=None):
        url = self._parse_url(url)
        headers = [
            ("Host", url.netloc),
            ("User-Agent", h11.PRODUCT_ID),
            ("Accept", "application/json"),
            ("Connection", "close"),
        ]
        if body is not None:
            headers += [
                ("Content-Type", content_type),
                ("Content-Length", str(len(body))),
            ]

        conn = h11.Connection(our_role=h11.CLIENT)

        to_send = conn.send(
            h11.Request(method=method, target=url.path or "/", headers=headers)
        )
        if body is not None:
            to_send += conn.send(h11.Data(body))
        to_send += conn.send(h11.EndOfMessage())

        async with self._connect(url) as (reader, writer):
            writer.write(to_send)
            await writer.drain()

            async def receive_all():
                error_status = None
                while True:
                    event = conn.next_event()
                    if event is h11.NEED_DATA:
                        conn.receive_data(await reader.read(4096))
                    elif isinstance(event, h11.Response):
                        if event.status_code != 200:
                            error_status = event.status_code
                            error_data = b""
                    elif isinstance(event, h11.Data):
                        if error_status is None:
                            yield event.data
                        else:
                            error_data += event.data
                    elif isinstance(event, (h11.EndOfMessage, h11.PAUSED)):
                        break

                if error_status is not None:
                    raise ProtocolError(
                        f"Expected status code 200, got {error_status}: {error_data}"
                    )

            content = b"".join([data async for data in receive_all()])

        return json.loads(content)

    @contextlib.asynccontextmanager
    async def _connect(self, url):
        proxy = self._get_proxy(url)
        if proxy:
            tunnel = self.open_tunnel(
                proxy.host, proxy.port, url.host, url.port, ssl_=proxy.ssl
            )
            async with tunnel as (reader, writer, trailing_data):
                if trailing_data:
                    raise ProtocolError("Unexpected trailing_data")
                if url.ssl:
                    await _start_tls(writer, url)
                yield reader, writer
        else:
            async with util.connect(url.host, url.port, ssl=url.ssl) as (
                reader,
                writer,
            ):
                yield reader, writer

    def _get_proxy(self, url):
        proxy = self._proxies.get(url.scheme)
        if proxy and not is_proxy_disabled(url.host, self._proxies.get("no", "")):
            return proxy
        return None

    def _parse_url(self, url):
        url = urllib.parse.urlsplit(url)
        if url.scheme == "https":
            default_port = 443
            ssl_ = self._ssl_ctx
        elif url.scheme == "http":
            default_port = 80
            ssl_ = False
        else:
            raise ValueError(f'Unknown scheme "{url.scheme}"')

        port = url.port or default_port

        if not url.hostname:
            raise ValueError(f'No hostname in URL "{url}"')

        return _ParsedURL(
            url.scheme,
            url.netloc,
            url.hostname,
            port,
            url.path,
            url.query,
            url.fragment,
            url.username,
            url.password,
            ssl_,
        )

    @contextlib.asynccontextmanager
    async def open_tunnel(
        self, proxy_host, proxy_port, target_host, target_port, ssl_=False
    ):
        headers = [
            ("Host", f"{target_host}:{target_port}"),
            ("User-Agent", h11.PRODUCT_ID),
        ]

        conn = h11.Connection(our_role=h11.CLIENT)
        to_send = conn.send(
            h11.Request(
                method="CONNECT", target=f"{target_host}:{target_port}", headers=headers
            )
        )

        async with util.connect(proxy_host, proxy_port, ssl=ssl_) as (reader, writer):
            writer.write(to_send)
            writer.write(conn.send(h11.EndOfMessage()))
            await writer.drain()

            while True:
                event = conn.next_event()
                if event is h11.NEED_DATA:
                    conn.receive_data(await reader.read(4096))
                elif isinstance(event, h11.Response):
                    response = event
                    if response.status_code != 200:
                        raise ProtocolError(
                            f"Expected status code 200, got {event.status_code}"
                        )

                    yield reader, writer, conn.trailing_data[0]
                    break
                else:
                    raise ProtocolError(f"Unexpected event: {event}")

    @contextlib.asynccontextmanager
    async def websocket(self, url, auth=None):
        url = self._parse_url(url)

        ws_scheme = "ws" if url.scheme == "http" else "wss"
        ws_uri = f"{ws_scheme}://{url.netloc}{url.path}"
        ws_uri = websockets.uri.parse_uri(ws_uri)

        protocol = websockets.ClientProtocol(ws_uri)

        async with self._connect(url) as (reader, writer):
            async with _WebsocketConnection(protocol, reader, writer, auth) as con:
                yield con


class _WebsocketConnection:
    _close_timeout = 10

    def __init__(self, protocol, reader, writer, auth=None):
        self._protocol = protocol
        self._reader = reader
        self._writer = writer
        self._chunks = []
        self._decoder = None
        self._send_lock = asyncio.Lock()
        self._auth = auth

    async def __aenter__(self):
        request = self._protocol.connect()
        if self._auth:
            request.headers["Authorization"] = self._auth

        # sending handshake request
        self._protocol.send_request(request)
        await self._send_protocol_data()

        # wait for handshake response
        self._iterator = self._receive()
        await self._iterator.__anext__()
        logger.debug("WebSocket connection established")

        return self

    async def __aexit__(self, exc_type, exc, tb):
        pass

    async def send(self, data):
        if isinstance(data, str):
            async with self._send_context():
                self._protocol.send_text(data.encode())
        elif isinstance(data, (bytes, bytearray, memoryview)):
            async with self._send_context():
                self._protocol.send_binary(data)
        else:
            raise TypeError("data can be either a string or bytes")

    async def receive_iter(self):
        async for message in self._iterator:
            yield message

    async def _receive(self):
        deadline = None
        while True:
            async with util.timeout_at(deadline) as timeout:
                data = await self._reader.read(4096)
                if data:
                    self._protocol.receive_data(data)

                    if self._protocol.close_expected():
                        logger.debug("Close expected, setting connection timeout")
                        now = asyncio.get_running_loop().time()
                        deadline = now + self._close_timeout
                        timeout.reschedule(deadline)

                    # handle pings
                    async with self._send_lock:
                        await self._send_protocol_data()

                    for message in self._handle_event():
                        yield message
                else:
                    logger.debug("Server closed connection")
                    self._protocol.receive_eof()
                    # maybe half close connection
                    async with self._send_lock:
                        await self._send_protocol_data()
                    raise self._protocol.close_exc

    def _handle_event(self):
        for event in self._protocol.events_received():
            if isinstance(event, websockets.frames.Frame):
                yield from self._defragment(event)
            elif isinstance(event, websockets.http11.Response):
                logger.debug("Received WebSocket handshake response: %s", event)
                if self._protocol.handshake_exc is not None:
                    raise self._protocol.handshake_exc
                yield

    def _defragment(self, frame):
        if frame.opcode is websockets.frames.Opcode.TEXT:
            self._decoder = UTF8Decoder(errors="strict")
        elif frame.opcode is websockets.frames.Opcode.BINARY:
            self._decoder = None
        elif frame.opcode is websockets.frames.Opcode.CONT:
            pass
        else:
            return

        if self._decoder is not None:
            data = self._decoder.decode(frame.data, final=frame.fin)
        else:
            data = frame.data

        self._chunks.append(data)

        if frame.fin:
            joiner = b"" if self._decoder is None else ""
            message = joiner.join(self._chunks)
            self._chunks.clear()
            yield message

    @contextlib.asynccontextmanager
    async def _send_context(self):
        async with self._send_lock:
            try:
                yield
            finally:
                await self._send_protocol_data()

    async def _send_protocol_data(self):
        for data in self._protocol.data_to_send():
            if data:
                self._writer.write(data)
                await self._writer.drain()
            else:
                if self._writer.can_write_eof():
                    self._writer.write_eof()


@dataclass
class _ParsedURL:
    scheme: str
    netloc: str
    host: str
    port: int
    path: str
    query: str
    fragment: str
    username: Optional[str]
    password: Optional[str]
    ssl: Union[bool, ssl.SSLContext]


def is_proxy_disabled(host, no_proxy_env):
    if not host or not no_proxy_env:
        return False

    if no_proxy_env == "*":
        return True

    def patterns(network_type=None):
        for pattern in no_proxy_env.split(","):
            pattern = pattern.strip()
            if pattern:
                if network_type is not None:
                    try:
                        pattern = network_type(pattern, strict=False)
                    except ValueError:
                        continue
                yield pattern

    is_disabled = False

    if host[0] == "[":
        # match IPv6
        return _is_proxy_disabled_ipv6(host, patterns(ipaddress.IPv6Network))
    else:
        try:
            addr = ipaddress.IPv4Address(host)
        except ValueError:
            # neither IPv4 nor IPv6 address, match hostname
            is_disabled = _is_proxy_disabled_host(host, patterns())
        else:
            # match IPv4
            for net in patterns(ipaddress.IPv4Network):
                if addr in net:
                    is_disabled = True
                    break

    return is_disabled


def _is_proxy_disabled_ipv6(host, disabled_networks):
    end = host.find("]")
    if end > 0:
        try:
            addr = ipaddress.IPv6Address(host[1:end])
        except ValueError:
            pass
        else:
            for net in disabled_networks:
                if addr in net:
                    return True
    return False


def _is_proxy_disabled_host(host, patterns):
    # ignore trailing dots in the host name
    if host[-1] == ".":
        host = host[:-1]

    # ignore case
    host = host.lower()

    for pattern in patterns:
        # ignore trailing dots in the pattern to check
        if pattern[-1] == ".":
            pattern = pattern[:-1]

        if pattern and pattern[0] == ".":
            # ignore leading pattern dot as well
            pattern = pattern[1:]

        if not pattern:
            continue

        # exact match: example.com matches 'example.com'
        if host == pattern.lower():
            return True

        # tail match: www.example.com matches 'example.com'
        # note: nonexample.com does not match 'example.com'
        if host.endswith(f".{pattern}"):
            return True

    return False


# pylint: disable=protected-access
# remove, if Python version < 3.11 is no longer supported
async def _start_tls(writer, url):
    if hasattr(writer, "start_tls"):
        await writer.start_tls(url.ssl, server_hostname=url.host)
    else:
        # backported from 3.11, commit 6217864f ("gh-79156: Add start_tls()
        # method to streams API (#91453)")
        protocol = writer._protocol
        await writer.drain()
        loop = asyncio.get_running_loop()
        new_transport = await loop.start_tls(
            writer._transport, protocol, url.ssl, server_hostname=url.host
        )
        writer._transport = new_transport

        protocol._stream_writer = writer
        protocol._transport = new_transport
        protocol._over_ssl = True



from typing import Dict, List

import pydantic

UsbId = pydantic.constr(regex=r"^[1-9][0-9]*-[1-9][0-9]*(\.[1-9][0-9]*)*$")


class UsbImportDesc(pydantic.BaseModel):
    port_num: pydantic.NonNegativeInt


class TcpImportDesc(pydantic.BaseModel):
    local_port: pydantic.PositiveInt


class ImportedPart(pydantic.BaseModel):
    compatible: List[str]
    usb: Dict[str, UsbImportDesc] = {}
    tcp: Dict[str, TcpImportDesc] = {}


class ImportDesc(pydantic.BaseModel):
    name: str
    parts: Dict[str, ImportedPart]


class UsbExportDesc(pydantic.BaseModel):
    usbid: UsbId


class TcpExportDesc(pydantic.BaseModel):
    host: str
    port: pydantic.PositiveInt


class ExportedPart(pydantic.BaseModel):
    compatible: List[str]
    usb: Dict[str, UsbExportDesc] = {}
    tcp: Dict[str, TcpExportDesc] = {}


class ExportDesc(pydantic.BaseModel):
    port: pydantic.PositiveInt
    parts: List[ExportedPart]


class Place(ExportDesc):
    id: pydantic.PositiveInt
    # host: pydantic.IPvAnyAddress
    #   can't serialize IP address with json.dumps()
    #   TODO: maybe drop pydantic as a dependency
    host: str



#!/usr/bin/env python3

import asyncio
import contextvars
import dataclasses
import functools
import itertools
import json
import logging
import textwrap
import traceback
from typing import Any, Optional, Union

import not_my_board._util as util

logger = logging.getLogger(__name__)
channel_var = contextvars.ContextVar("channel")


CODE_INTERNAL_ERROR = -32603
CODE_INVALID_REQUEST = -32600
CODE_METHOD_NOT_FOUND = -32601


class Channel(util.ContextStack):
    """Send and receive remote procedure calls with JSON RPC

    `send` is an async function that takes messages (bytes) and sends it over
    the wire.
    `receive_iter` is an async iterator, that yields messages (bytes) whenever
    they are available.
    The `api_obj` is used to handle incoming calls. If it has a coroutine with
    a name, that matches with the received method request, then it is called.
    The return value is sent as a response. If the method throws an exception,
    then an error with the traceback is sent to the remote caller.

    To call methods on the remote side, just call a coroutine method with the
    same name on the channel object. The call is automatically converted into a
    remote procedure call. It waits for the result and converts it to a return
    value. If an error is returned, then it is converted back into an
    exception. If the call is canceled, then the remote execution is also
    canceled.
    """

    def __init__(self, send, receive_iter, api_obj=None):
        self._send = send
        self._receive_iter = receive_iter
        self._api_obj = api_obj
        self._tasks = set()
        self._request_tasks_by_id = {}
        self._id_generator = itertools.count(start=1)
        self._pending = {}

        # TODO should be False before communicate_forever() is running
        self._is_receiving = True

    def set_api_object(self, api_obj):
        """Set the API object, that is used to handle incoming calls."""
        self._api_obj = api_obj

    async def communicate_forever(self):
        """This function needs to run in a task, while the channel is used"""
        channel_var.set(self)
        try:
            async for raw_data in self._receive_iter:
                try:
                    await self._receive(raw_data)
                except Exception:
                    traceback.print_exc()
        finally:
            self._is_receiving = False
            for future in self._pending.values():
                if not future.done():
                    future.set_exception(RuntimeError("Connection closed"))
            await util.cancel_tasks(self._tasks.copy())

    async def _context_stack(self, stack):
        bg_task = util.background_task(self.communicate_forever())
        await stack.enter_async_context(bg_task)

    def __getattr__(self, method_name):
        if method_name.startswith("_"):
            raise AttributeError(f"invalid attribute '{method_name}'")
        return functools.partial(self._call, method_name)

    async def _receive(self, raw_data):
        info = {"id": None, "is_request": False}
        try:
            message = _parse_message(raw_data, info)
        except Exception as e:
            if info["id"] is not None:
                if info["is_request"]:
                    response = ErrorResponse.with_traceback(
                        info["id"], CODE_INVALID_REQUEST, "Invalid Request"
                    )
                    await self._send(bytes(response))
                    return
                else:
                    future = self._pending.get(info["id"])
                    if future and not future.done():
                        future.set_exception(e)
                        return

            raise

        if isinstance(message, Request):
            task = asyncio.create_task(self._handle_request(message))
            self._tasks.add(task)
            task.add_done_callback(self._tasks.discard)
        elif isinstance(message, Response):
            await self._handle_response(message)
        else:  # ErrorResponse
            await self._handle_error_response(message)

    async def _handle_request(self, request):
        next_error = (CODE_METHOD_NOT_FOUND, "Method not found")
        try:
            if request.method == "rpc.cancel":
                method = self._cancel_local
            else:
                if request.method.startswith("_"):
                    raise ProtocolError(f'method "{request.method}" not allowed')

                method = getattr(self._api_obj, request.method)

                if getattr(method, "_jsonrpc_hidden", False):
                    raise ProtocolError(f'method "{request.method}" is marked hidden')

                logger.info("Method call: %s", request.method)

            next_error = CODE_INTERNAL_ERROR, None
            if request.id is not None:
                try:
                    self._request_tasks_by_id[request.id] = asyncio.current_task()
                    result = await method(*request.args, **request.kwargs)
                    response = Response(request.id, result)
                    await self._send(bytes(response))
                finally:
                    del self._request_tasks_by_id[request.id]
            else:
                await method(*request.args, **request.kwargs)
        except Exception as e:
            if request.id is not None:
                code, message = next_error
                if message is None:
                    message = str(e)
                response = ErrorResponse.with_traceback(request.id, code, message)
                await self._send(bytes(response))
            else:
                traceback.print_exc()

    async def _cancel_local(self, id_):
        if id_ in self._request_tasks_by_id:
            await util.cancel_tasks([self._request_tasks_by_id[id_]])

    async def _handle_response(self, response):
        future = self._pending.get(response.id)
        if future and not future.done():
            future.set_result(response.result)

    async def _handle_error_response(self, error_response):
        exc = error_response.as_exception()

        future = self._pending.get(error_response.id)
        if future and not future.done():
            future.set_exception(exc)
        else:
            raise exc

    async def _call(self, method_name, *args, **kwargs):
        if not self._is_receiving:
            raise RuntimeError("Channel communication is already closed")

        if kwargs.pop("_notification", False):
            id_ = None
        else:
            id_ = next(self._id_generator)

        if args and kwargs:
            raise RuntimeError("Use either args or kwargs")

        request = Request(id_, method_name, args or kwargs)
        logger.info("Calling: %s", request.method)

        if id_ is not None:
            return await self._send_request(request)
        else:
            # send notification
            await self._send(bytes(request))

    async def _send_request(self, request, send_cancellation=True):
        future = asyncio.get_running_loop().create_future()
        self._pending[request.id] = future
        try:
            await self._send(bytes(request))
            return await self._pending[request.id]
        except asyncio.CancelledError:
            if send_cancellation:
                logger.info("Canceling: %s", request.method)
                await self._cancel_remote(request.id)
            raise
        finally:
            del self._pending[request.id]

    async def _cancel_remote(self, to_cancel_id):
        id_ = next(self._id_generator)
        request = Request(id_, "rpc.cancel", [to_cancel_id])

        # don't request to cancel the cancellation if this task is canceled
        await self._send_request(request, send_cancellation=False)


def hidden(func):
    """Decorator to mark a function as hidden.

    Hidden functions can't be called by the remote. Use this if you want to
    keep it public for local users, otherwise just use a leading underscore in
    the method name.
    """
    # pylint: disable=protected-access
    func._jsonrpc_hidden = True
    return func


def get_current_channel():
    return channel_var.get()


def _parse_message(raw_data, info):
    data = json.loads(raw_data)
    id_ = data.get("id")
    if id_ is not None:
        if not isinstance(id_, (str, int)):
            raise ProtocolError('"id" must be a string or number')
        info["id"] = id_

    # check if it is a Request
    method = data.get("method")
    if method is not None:
        info["is_request"] = True

        if not isinstance(method, str):
            raise ProtocolError('"method" must be a string')

        params = data.get("params", [])
        if not isinstance(params, (list, dict)):
            raise ProtocolError('"params" must be a structured value')

        return Request(id_, method, params)

    # must be a Response or ErrorResponse
    if id_ is None:
        raise ProtocolError('"id" is required')

    # check if it is an ErrorResponse
    error = data.get("error")
    if error:
        code = error["code"]
        if not isinstance(code, int):
            raise ProtocolError('"error.code" must be an integer')

        message = error["message"]
        if not isinstance(message, str):
            raise ProtocolError('"error.message" must be a string')

        filtered_error = {
            "code": code,
            "message": message,
            "data": error.get("data"),
        }
        return ErrorResponse(id_, filtered_error)

    # must be a Response
    return Response(id_, data["result"])


@dataclasses.dataclass
class _Message:
    jsonrpc: str = dataclasses.field(default="2.0", init=False)
    id: Optional[Union[int, str]]

    def __bytes__(self):
        body = {}
        for field in dataclasses.fields(self):
            if field.name == "id" and self.id is None:
                # skip optional id
                continue

            body[field.name] = getattr(self, field.name)

        return json.dumps(body).encode()


@dataclasses.dataclass
class Request(_Message):
    method: str
    params: Union[list, dict]

    @property
    def args(self):
        if isinstance(self.params, list):
            return self.params
        return []

    @property
    def kwargs(self):
        if isinstance(self.params, dict):
            return self.params
        return {}


@dataclasses.dataclass
class Response(_Message):
    result: Any


@dataclasses.dataclass
class ErrorResponse(_Message):
    error: dict

    @classmethod
    def with_traceback(cls, id_, code, message):
        error = {
            "code": code,
            "message": message,
            "data": {
                "traceback": traceback.format_exc(),
            },
        }
        return cls(id_, error)

    def as_exception(self):
        return RemoteError(
            self.error["code"], self.error["message"], self.error["data"]
        )


class RemoteError(Exception):
    def __init__(self, code, message, data):
        if isinstance(data, dict) and "traceback" in data:
            details = textwrap.indent(data["traceback"], " " * 4).rstrip()
            super().__init__(f"{message}\n{details}")
        else:
            super().__init__(message)
        self.code = code
        self.data = data


class ProtocolError(Exception):
    pass
